"""Option classes."""

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union


def _negate_description(description: str) -> str:
    """Negate a description.

    This is extremely crude, but maybe it is good enough.
    """
    # Decapitalize it.
    lower_description = f"{description[0].lower()}{description[1:]}"
    return f"Don't {lower_description}"


def generate_multi_help_statement(
    varname: str,
    choices: List[Tuple[Union[str, int, bool], str]],
    default: Union[str, int, bool],
) -> List[Tuple[str, str]]:
    """Generate a help statement for a multiple choice option."""
    is_default = "* "
    is_not_default = "  "

    options = []
    for choice in choices:
        choiceval, choicedescr = choice
        if choiceval == default:
            options.append((f"{is_default}{varname}={choiceval}", choicedescr))
        else:
            options.append((f"{is_not_default}{varname}={choiceval}", choicedescr))

    return options


class Option:
    """Base option class."""

    title: str
    default: Union[str, int, bool]
    _ispatchoption = True

    def __init__(self, value: Union[str, int, bool]) -> None:
        """Initialize option with value."""
        self.value: Union[str, int, bool] = value

    @classmethod
    def asdict(cls, varname: str) -> Dict[str, Any]:
        """Return as dictionary."""
        return {
            "name": varname,
            "title": cls.title,
        }


class BoolOption(Option):
    """Boolean option."""

    def get_value(self) -> bool:
        """Return value."""
        if isinstance(self.value, str):
            if self.value.lower() in ("false", "off", "no"):
                return False
            elif self.value.lower() in ("true", "on", "yes"):
                return True

        return bool(self.value)

    @classmethod
    def from_default(cls) -> "BoolOption":
        """Instantiate with default value."""
        return cls(cls.default)

    @classmethod
    def help(cls, varname: str) -> List[Tuple[str, str]]:
        """Generate a help text."""
        choices = [(True, cls.title), (False, _negate_description(cls.title))]
        return generate_multi_help_statement(varname, choices, cls.default)


class EnumOption(Option):
    """Enumeration option."""

    enum: List[Tuple[str, str]]

    def __init__(self, value: Union[str, int, bool]) -> None:
        """Validate input."""
        choices = [choice[0] for choice in self.enum]
        if str(value) not in choices:
            raise ValueError(f"Unsupported choice: {value}")
        super().__init__(value)

    def get_value(self) -> str:
        """Return value."""
        return str(self.value)

    @classmethod
    def from_default(cls) -> "EnumOption":
        """Instantiate with default value."""
        return cls(cls.default)

    @classmethod
    def asdict(cls, varname: str) -> Dict[str, Any]:
        """Return as dictionary."""
        base = super().asdict(varname)
        base["enum"] = cls.enum
        return base

    @classmethod
    def help(cls, varname: str) -> List[Tuple[str, str]]:
        """Generate a help text."""
        # Coerce choices to make mypy happy
        choices: List[Tuple[Union[str, int, bool], str]] = [
            choice for choice in cls.enum
        ]
        return generate_multi_help_statement(varname, choices, cls.default)


class NumberOption(Option):
    """NUmber option."""

    number: Tuple[int, int]

    def __init__(self, value: Union[str, int, bool]) -> None:
        """Validate input."""
        if isinstance(value, str):
            if not value.isdecimal():
                raise ValueError(f"Expected a decimal number, but got {value}.")

        super().__init__(value)

    def get_value(self) -> int:
        """Return value."""
        return int(self.value)

    @classmethod
    def help(cls, varname: str) -> List[Tuple[str, str]]:
        """Generate a help statement."""
        # We won't list every number value, so we specify the options as a range instead.
        return [(f"  {varname}={cls.number[0]}..{cls.number[1]}", cls.title)]

    @classmethod
    def asdict(cls, varname: str) -> Dict[str, Any]:
        """Return as dictionary."""
        base = super().asdict(varname)
        base["number"] = cls.number
        return base

    @classmethod
    def from_default(cls) -> "NumberOption":
        """Instantiate with default value."""
        return cls(cls.default)


def option(
    description: str,
    section: str = "Options",
    *,
    choices: Optional[List[Tuple[str, str]]] = None,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    default: Optional[Union[str, int, bool]] = None,
) -> Any:
    """Define a configurable variable.

    Args:
        description: Description of the variable
        section: The name of the configuration pane where this option should be shown.

    A number of different option types can be generated. Without any further
    options, the option will be a boolean true/false.

    Number Option:
        minimum: integer
        maximum: integer

    Enumeration Option:
        choices: List of tuples. The Tuples define (value, description)

    Default:
        You can specify a default, otherwise it will be derived as follows:
        Bool: True
        Number: Lowest number
        Enum: First value.

    """
    params: Dict[str, Any] = {
        "title": description,
        "section": section,
    }
    bases: Tuple[type, ...]
    if default:
        params["default"] = default

    if choices:
        # Slightly hacky. We construct a dynamic child class that sets the given
        # title and section. This allows us to use it as a string.
        params["enum"] = choices
        if not default:
            params["default"] = choices[0][0]
        bases = (EnumOption, Option)
        return type("LocalEnumOption", bases, params)  # type: ignore
    elif minimum is not None and maximum is not None:
        params["number"] = (minimum, maximum)
        if not default:
            params["default"] = minimum
        bases = (NumberOption, Option)

        return type("LocalNumberOption", bases, params)  # type: ignore
    else:
        bases = (BoolOption, Option)
        if not default:
            params["default"] = True
        return type("LocalBoolOption", bases, params)  # type: ignore


@dataclass
class Preset:
    """Preset class."""

    title: str
    parameters: Dict[str, Any]

    def asdict(self, name: str) -> Dict[str, Any]:
        """Construct preset dictionary."""
        return {"name": name, "title": self.title, "parameters": self.parameters}


def preset(title: str, parameters: Dict[str, Any]) -> Preset:
    """Define a preset.

    Args:
        title: The name of the preset.
        parameter: A dictionary containing parameter to value mappings.
    """
    return Preset(title, parameters)



"""Utility functions."""

import textwrap
from typing import Optional

COMMENT_LINE_LENGTH = 88


def write_patch_section(name: str, comment: Optional[str] = None) -> str:
    """Write a patch section."""
    delim = f"# {'-'*49}"
    if not comment:
        comment = ""
    else:
        comment = format_comment(comment) + "\n"
    return f"{delim}\n# {name}\n{comment}{delim}\n"


def format_comment(comment_str: str, dedent: bool = False) -> str:
    """Format comments."""
    if dedent:
        comment_str = textwrap.dedent(comment_str)
    lines = textwrap.wrap(comment_str, COMMENT_LINE_LENGTH)
    commented_lines = [f"# {line}" for line in lines]
    return "\n".join(commented_lines)



"""Class based patch generator interface."""

import argparse
import json
import textwrap

from abc import abstractmethod
from itertools import groupby
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
)

from .base import Circuit, Controller, Label
from .options import Option, BoolOption, NumberOption, EnumOption, Preset
from .utils import write_patch_section

DEFAULT_SECTION_NAME = "Generated Patch"


def _split_params(params: List[str]) -> Dict[str, str]:
    """Split input parameters."""
    patch_params: Dict[str, str] = {}
    for param in params:
        if "=" not in param:
            raise ValueError("Parameters must be input as key=value")
        key, value = param.split("=")
        if key in patch_params:
            raise ValueError(f"Duplicate key {key} found.")
        # We convert everything to strings.
        patch_params[key] = str(value)

    return patch_params


class MetaMetaPatch(type):
    """Metaclass for metapatches."""

    title: str
    description: str

    @property
    def synopsis(cls) -> Dict[str, Any]:
        """Get synopsis."""
        sections: Dict[str, Dict[str, Option]] = {}
        presets: List[Dict[str, Any]] = []
        for name, opt in cls.__dict__.items():
            if hasattr(opt, "_ispatchoption"):
                if opt.section not in sections:
                    sections[opt.section] = {}
                sections[opt.section][name] = opt
            elif isinstance(opt, Preset):
                presets.append(opt.asdict(name))

        section_list: List[Dict[str, Any]] = []
        for section_name, options in sections.items():
            subsection: Dict[str, Any] = {"title": section_name, "options": []}
            for option_name, option in options.items():
                subsection["options"].append(option.asdict(option_name))
            section_list.append(subsection)

        synopsis = {
            "title": cls.title,
            "description": cls.description,
            "sections": section_list,
            "presets": presets,
        }
        return synopsis

    @property
    def presets(cls) -> List[Dict[str, Any]]:
        """Fetch presets."""
        presets: List[Dict[str, Any]] = []
        for name, opt in cls.__dict__.items():
            if isinstance(opt, Preset):
                presets.append(opt.asdict(name))

        return presets


class PatchGenerator(metaclass=MetaMetaPatch):
    """Patch Generator."""

    title: str
    description: str
    __patchvars__: List[str]

    def __new__(cls, **_: Any) -> "PatchGenerator":
        """Build class."""
        cls.__patchvars__ = []
        for param, opt in cls.__dict__.items():
            if hasattr(opt, "_ispatchoption"):
                cls.__patchvars__.append(param)

        return super().__new__(cls)

    def __init__(self, **kwargs: str) -> None:
        """Initialize class."""
        # Set all options not defined in the input to their defaults.
        for opt in self.__patchvars__:
            if opt not in kwargs:
                var = getattr(self, opt)
                assert issubclass(var, (BoolOption, EnumOption, NumberOption))
                setattr(self, opt, var.from_default().get_value())

        for key, value in kwargs.items():
            if key in self.__patchvars__:
                var = getattr(self, key)
                assert issubclass(var, (BoolOption, EnumOption, NumberOption))
                setattr(self, key, var(value).get_value())
            else:
                raise ValueError(f"Unknown parameter {key}")

        self.circuits: List[Circuit] = []
        self.controllers: List[Controller] = []
        self.labels: List[Label] = []
        self._section: Optional[str] = None
        # This dictionary stores comments for sections.
        self.sections: Dict[str, Optional[str]] = {}

    @classmethod
    def load_preset(cls, preset_name: str) -> "PatchGenerator":
        """Load preset."""
        preset = getattr(cls, preset_name)
        if preset:
            return cls(**preset.parameters)
        raise ValueError(f"Uknown preset: {preset_name}")

    @classmethod
    def _handle_cli(cls) -> argparse.Namespace:
        """Handle CLI."""
        parser = argparse.ArgumentParser(
            f'DROID patch generator "{cls.title}"',
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        parser.description = cls.description
        parser.add_argument(
            "parameters",
            nargs="*",
            help="Patch generator parameters.",
            metavar="param=value ...",
        )
        parser.add_argument(
            "--synopsis",
            "-s",
            action="store_true",
            help="Output possible parameters as JSON",
        )
        parser.add_argument(
            "-p", "--preset", metavar="P", help="Use settings from preset P"
        )
        presets = cls.presets
        presets_str = "\n".join(
            [f"{preset['name']:<12} {preset['title']:<12}" for preset in presets]
        )
        epilog = f"Available Presets:\n{presets_str}\n\n"
        params = "parameters (defaults are marked with *):\n\n"
        all_param_help = {}
        for key, value in cls.__dict__.items():
            if not hasattr(value, "_ispatchoption"):
                continue
            if issubclass(value, (BoolOption, EnumOption, NumberOption)):
                all_param_help[value.title] = value.help(key)

        if all_param_help:
            longest = max(
                [
                    len(varname[0])
                    for param in list(all_param_help.values())
                    for varname in param
                ]
            )
            for title, opthelp in all_param_help.items():
                params += textwrap.indent(title, "  ") + "\n"
                for opt, descr in opthelp:
                    params += textwrap.indent(f"{opt:<{longest}}  {descr}", "    ")
                    params += "\n"

                params += "\n"

            epilog += params
            parser.epilog = epilog

        args = parser.parse_args()

        return args

    @classmethod
    def run(cls) -> None:
        """Process parameters given via command line arguments."""
        args = cls._handle_cli()
        if args.synopsis:
            print(json.dumps(cls.synopsis))
            return
        if args.preset:
            patch = cls.load_preset(args.preset)
            print(str(patch))
            return

        patch_parameters: List[str] = args.parameters
        parameters = _split_params(patch_parameters)

        patch = cls(**parameters)
        print(str(patch))

    def add_circuit(
        self,
        name: str,
        params: Mapping[str, str],
        comment: Optional[str] = None,
    ) -> None:
        """Add a circuit.

        Args:
            name: Circuit name, e.g. copy for a [copy] circuit
            params: Dictionary of circuit parameters.
            comment: Optional comment for the circuit.
        """
        self.circuits.append(
            Circuit(name=name, parameters=params, comment=comment, section=self.section)
        )

    def add_label(
        self, item: str, short_label: str, long_label: Optional[str] = None
    ) -> None:
        """Add a label.

        Args:
            item: string of the thing you want to label, e.g. O1 or G1.2
            short_label: The short label
            long_label: Optional longer label
        """
        self.labels.append(Label.from_item(item, short_label, long_label))

    def _generate_labels(self) -> str:
        """Generate label strings."""
        sorted_labels = {}
        for label in self.labels:
            if label.heading not in sorted_labels:
                sorted_labels[label.heading] = []
            sorted_labels[label.heading].append(label)

        labels = ""
        for heading, list_labels in sorted_labels.items():
            labels += f"# {heading.upper()}:\n"
            for label in list_labels:
                labels += str(label)
            labels += "\n"

        return labels

    def add_controller(self, type: str, position: int) -> None:
        """Add a controller at a given position.

        Args:
            type: Type of controller, e.g. B32
            position: controller position, e.g. 1
        """
        self.controllers.append(Controller(type, position))

    @property
    def section(self) -> Optional[str]:
        """Get section."""
        return self._section

    @section.setter
    def section(self, name: str) -> None:
        """Set the section."""
        if not isinstance(name, str):
            raise ValueError("Section name must be a string.")
        if name not in self.sections:
            self.sections[name] = None
        self._section = name

    def add_section(self, name: str, comment: Optional[str] = None) -> None:
        """Add a section with an optional comment."""
        self._section = name
        if name not in self.sections or self.sections.get(name) != comment:
            self.sections[name] = comment

    @property
    def _has_sections(self) -> bool:
        """Check if sections are defined."""
        sections = [circuit.section for circuit in self.circuits]
        return any(sections)

    def _get_circuits_as_strings(self) -> str:
        """Get circuits as a string.

        This is for circuits when there are no sections at all.
        """
        circuits = [str(circuit) for circuit in self.circuits]
        return "\n".join(circuits)

    def _get_circuits_with_sections(self) -> str:
        """Get circuits as a string with sections."""
        circuits = []

        groups = {}
        for section, grouped_circuits in groupby(self.circuits, lambda x: x.section):
            if section is None:
                section_name = DEFAULT_SECTION_NAME
            else:
                section_name = section
            if section_name not in groups:
                groups[section_name] = []
            groups[section_name].extend(grouped_circuits)

        for section_name, grouped_circuits in groups.items():
            if section_name == DEFAULT_SECTION_NAME:
                circuits.append(write_patch_section(DEFAULT_SECTION_NAME))
            else:
                comment = self.sections.get(section_name)
                circuits.append(write_patch_section(section_name, comment))

            for circuit in grouped_circuits:
                circuits.append(str(circuit))

        return "\n".join(circuits)

    def __str__(self) -> str:
        """Output patch as string."""
        self._generate()

        labels = self._generate_labels()
        sorted_controllers = sorted(self.controllers, key=lambda x: x.position)
        controllers = "\n".join([str(controller) for controller in sorted_controllers])
        if self._has_sections:
            circuits = self._get_circuits_with_sections()
        else:
            circuits = self._get_circuits_as_strings()

        return f"{labels}{controllers}\n\n{circuits}"

    def _generate(self) -> None:
        """Generate patch.

        This is an internal function that ensures that we reset the patch if needed.
        """
        if not self.circuits and not self.controllers:
            self.generate()

    @abstractmethod
    def generate(self) -> None:
        """Generate patch.

        Define this in your subclass, and use it to add circuits and controllers.
        """



"""Base classes."""

import re
from dataclasses import dataclass, field
from typing import Optional, Mapping

from .utils import format_comment, write_patch_section


@dataclass
class Circuit:
    """Circuit container class."""

    name: str
    parameters: Mapping[str, str] = field(default_factory=dict)
    comment: Optional[str] = None
    section: Optional[str] = None

    def __str__(self) -> str:
        """Output circuit as patch."""
        indent = 4 * " "
        comment = ""
        if self.comment:
            comment = format_comment(self.comment) + "\n"
        params = "\n".join(
            [f"{indent}{key} = {value}" for key, value in self.parameters.items()]
        )
        patch = f"{comment}[{self.name}]\n{params}\n"
        return patch


@dataclass
class Controller:
    """Droid controller container class."""

    type: str
    position: int

    def __str__(self) -> str:
        """Generate controller string."""
        return f"[{self.type}]"


@dataclass
class Section:
    """Section class."""

    name: str
    comment: Optional[str] = None

    def __str__(self) -> str:
        """Return section string."""
        output = ""
        if self.comment:
            output += format_comment(self.comment)

        output += write_patch_section(self.name)
        return output


@dataclass
class Label:
    """Label class."""

    heading: str
    item: str
    short_label: str
    long_label: Optional[str] = None

    def __str__(self) -> str:
        """Generate label.

        Does not include the heading.
        """
        label = f"#  {self.item}: [{self.short_label}]"
        if self.long_label:
            label += f" {self.long_label}"

        label += "\n"
        return label

    @classmethod
    def from_item(
        cls, item: str, short_label: str, long_label: Optional[str] = None
    ) -> "Label":
        """Construct from item."""
        if item.startswith("I"):
            # Input label
            section = "inputs"
        elif item.startswith("O"):
            section = "outputs"
        elif res := re.match(r"([A-Z])(\d+)\.(\d+)", item):
            module_num = res.group(2)
            controller_type = res.group(1)
            if controller_type == "G":
                if "." in item:
                    section = f"gates on module {module_num}"
                else:
                    section = "gates on x7"
            else:
                section = f"controller {module_num}"

        return cls(section, item, short_label, long_label)



#!/usr/bin/env python3
"""This is a small helper script that can be used to speed up some of the patch building."""
import argparse
import sys
import os
import re

from dataclasses import dataclass
from typing import Iterator, Dict, List, Optional


@dataclass
class Token:
    """Token container class."""

    type: str
    value: str
    line: int
    column: int


def tokenize(patch: str) -> Iterator[Token]:
    """Tokenize a patch."""
    spec = [
        ("CONTROLLER", r"(?<=\[)([a-z][0-9]+[a-z0-9]*)(?=\])"),
        ("CIRCUIT", r"(?<=\[)([a-z]+)(?=\])"),
        ("PARAM", r"([a-z0-9]+)\s*(?= =)"),
        ("VALUE", r"(?<= =\s)([^\s].*)(?=\n)"),
        ("NEWLINE", r"\n"),
        ("COMMENT", r"(?<=#\s)(.*)(?=\n)"),
    ]

    tok_regex = "|".join("(?P<%s>%s)" % pair for pair in spec)
    line_num = 1
    line_start = 0
    section_header = "-" * 49
    in_section = False
    section_title_derived = False
    for mo in re.finditer(tok_regex, patch):
        kind = mo.lastgroup
        value = mo.group()
        column = mo.start() - line_start
        if kind == "NEWLINE":
            line_start = mo.end()
            line_num += 1
            continue
        if kind == "COMMENT" and value == section_header:
            in_section = not in_section
            if in_section:
                section_title_derived = False
            continue
        if in_section:
            if kind == "COMMENT":
                if section_title_derived:
                    kind = "SECTION_COMMENT"
                else:
                    kind = "SECTION"
                    section_title_derived = True

        if kind is None:
            continue
        yield Token(kind, value, line_num, column)


def get_patch_stdin() -> str:
    """Read a patch from stdin."""
    buffer = []
    empty_lines = 0
    while True:
        line = sys.stdin.readline()
        if line == "\n":
            empty_lines += 1
            if empty_lines == 2:
                break
            else:
                buffer.append(line.rstrip("\n"))
        elif line == "":
            break
        else:
            buffer.append(line.rstrip("\n"))

    return "\n".join(buffer)


def _generate_section(name, comment: Optional[str], indent: bool = False) -> str:
    """Generate section."""
    indstr = ""
    c_indstr = " "
    level_start = 8
    level_cont = 12
    if indent:
        indstr = f"\n{' ' * level_start}"
        c_indstr = f"\n{' ' * level_cont}"
    if comment:
        comment = _format_comment(comment)
        return f'{indstr}self.add_section({c_indstr}"{name}",{c_indstr}"{comment}"{indstr})'

    return f'{indstr}self.add_section("{name}")'


def _params_to_str(params: Dict[str, str], start_indent: int = 4) -> str:
    """Convert params to string."""
    level_1 = start_indent
    level_2 = level_1 + 4
    indstr = " " * level_1
    c_indstr = " " * level_2

    text = "\n"
    text += indstr + "{"
    if not params:
        text += "}"
        return text
    text += "\n"
    for key, value in params.items():
        text += f'{c_indstr}"{key}": "{value}",\n'

    text += indstr + "}"
    return text


def _format_comment(comment: str) -> str:
    """Format comment.

    This escapes the newline characters.
    """
    return comment.replace("\n", "\\n")


def _generate_controller(name: str, index: int, indent: bool = False) -> str:
    """Generate a controller."""
    level_1 = 0
    if indent:
        level_1 = 8

    indstr = " " * level_1
    return f"{indstr}self.add_controller(\"{name}\", {index})"


def _generate_circuit(
    name: str, params: Dict[str, str], comment: Optional[str], indent: bool = False
) -> str:
    """Generate circuit."""
    level_1 = 0
    if indent:
        level_1 = 8

    level_2 = level_1 + 4

    indstr = f"\n{' ' * level_1}"
    c_indstr = f"\n{' ' * level_2}"
    param_str = _params_to_str(params, level_2)

    if comment:
        comment = _format_comment(comment)
        return f'{indstr}self.add_circuit({c_indstr}"{name}",{param_str},{c_indstr}"{comment}"{indstr})'
    return f'{indstr}self.add_circuit({c_indstr}"{name}",{param_str}{indstr})'


def generate_boilerplate(patch: str) -> str:
    """Generate full boilerplate."""
    boilerplate = """
import metapatch

class PatchGenerator(metapatch.PatchGenerator):
    \"\"\"Patch Generator boilerplate.

    Change this to something more meaningful.
    Also remember to add parameters and presets.
    \"\"\"
    title = "Title of your patch generator"  # CHANGEME
    description = "Description of your patch generator"  # CHANGEME

    def generate(self) -> None:
        \"\"\"Patch generator function.

        Modify this function to create your patch generator.
        \"\"\"
"""
    boilerplate += patch_to_metapatch(patch, indent=True)

    boilerplate += """
if __name__ == "__main__":
    PatchGenerator.run()
    """
    return boilerplate


def patch_to_metapatch(patch: str, indent: bool = False) -> str:
    """Convert a patch to a partial metapatch."""
    metapatch = []
    sections = {}
    circuits = []
    controllers = []
    curr_params = {}
    curr_param: Optional[str] = None
    curr_section: Optional[str] = None
    curr_circuit: Optional[str] = None
    curr_circuit_comments: List[str] = []
    for token in tokenize(patch):
        if token.type == "SECTION":
            curr_section = token.value
            sections[curr_section] = []
            continue
        elif token.type == "CONTROLLER":
            # Discard all previous comments
            controllers.append(token.value)
            curr_circuit_comments = []

        elif token.type == "CIRCUIT":
            if curr_circuit:
                circuits.append({curr_circuit: curr_params})
            curr_circuit = token.value
            curr_params = {}
            if curr_circuit_comments:
                curr_params["__comment"] = "\n".join(curr_circuit_comments)
                curr_circuit_comments = []
            if curr_section:
                curr_params["__section"] = curr_section

        elif token.type == "PARAM":
            curr_params[token.value] = ""
            curr_param = token.value
        elif token.type == "VALUE":
            curr_params[curr_param] = token.value

            curr_param = None
        elif token.type == "SECTION_COMMENT":
            sections[curr_section].append(token.value)
        elif token.type == "COMMENT":
            curr_circuit_comments.append(token.value)

    if curr_circuit:
        final_circuit = {curr_circuit: curr_params}
        if not circuits or final_circuit != circuits[-1]:
            circuits.append(final_circuit)

    for index, controller in enumerate(controllers, start=1):
        metapatch.append(_generate_controller(controller, index, indent))
    all_sections = [
        param.get("__section", "__NONE")
        for circuit in circuits
        for _, param in circuit.items()
    ]
    unique_sections = set(all_sections)
    for section in unique_sections:
        if section == "__NONE":
            # No section.
            section_circuits = [
                circ
                for circ in circuits
                for params in circ.values()
                if "__section" not in params
            ]
        else:
            section_comment = sections.get(section)
            if section_comment:
                section_comment = "\n".join(section_comment)
            metapatch.append(_generate_section(section, section_comment, indent))
            section_circuits = [
                circ
                for circ in circuits
                for params in circ.values()
                if params.get("__section") == section
            ]

        for circuit in section_circuits:
            for circuit_name, circuit_params in circuit.items():
                circuit_comment: Optional[str] = None
                if "__comment" in circuit_params:
                    circuit_comment = circuit_params["__comment"]
                    del circuit_params["__comment"]
                if "__section" in circuit_params:
                    del circuit_params["__section"]

                metapatch.append(
                    _generate_circuit(
                        circuit_name, circuit_params, circuit_comment, indent
                    )
                )

    return "\n".join(metapatch)


def cli() -> argparse.Namespace:
    """Parse cli."""
    parser = argparse.ArgumentParser(description="Metapatch helper script.")
    parser.add_argument(
        "infile",
        nargs="?",
        type=argparse.FileType("r", encoding="utf-8"),
        default=sys.stdin,
        help="Input file. Defaults to stdin.",
    )
    parser.add_argument(
        "outfile",
        nargs="?",
        type=argparse.FileType("w", encoding="utf-8"),
        default=sys.stdout,
        help="Output file. Defaults to stdout.",
    )
    parser.add_argument(
        "--boilerplate",
        action="store_true",
        help="Generate a full example patch generator.",
    )
    parser.epilog = (
        "If called without any arguments, you may input one or more droid circuits, "
        "and get the corresponding python code back."
    )
    args = parser.parse_args()
    return args


def main() -> None:
    """Main function."""
    args = cli()
    if args.infile.isatty():
        print("Paste a circuit configuration and finish with ctrl-d.\n")
    buffer = args.infile.read()
    if args.boilerplate:
        output = generate_boilerplate(buffer)
    else:
        output = patch_to_metapatch(buffer)

    args.outfile.write(output + os.linesep)


if __name__ == "__main__":
    main()
